package redner

import (
	"context"
	"errors"
	"fmt"

	"github.com/prometheus/client_golang/prometheus"

	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/utils"
)

var ErrEmptyPipeline = errors.New("empty pipeline")

// NewConversionPipeline returns a list of engine for converting a document of type 'fromType'
// to type 'toType'
// TODO do a real implementation, not this hard-coded hack.
func NewConversionPipeline(
	registry *rendering.EngineRegistry, fs *rendering.FS, fromType, toType string,
) (*ConversionPipeline, error) {
	if fromType == toType {
		return nil, ErrEmptyPipeline
	}

	for _, p := range paths {
		if p[0] == fromType && p[len(p)-1] == toType {
			steps := make([]ConversionStep, len(p)-1)

			from := p[0]
			for i, to := range p[1:] {
				engine, err := registry.FindEngine(from, to)
				if err != nil {
					return nil, err
				}

				steps[i] = ConversionStep{from, to, engine}
				from = to
			}

			return &ConversionPipeline{
				steps: steps,
				fs:    fs,
			}, nil
		}
	}

	engine, err := registry.FindEngine(fromType, toType)
	if err != nil {
		return nil, err
	}

	return &ConversionPipeline{
		steps: []ConversionStep{{fromType, toType, engine}},
		fs:    fs,
	}, nil
}

// ConversionStep is a step in a pipeline.
type ConversionStep struct {
	SourceType string
	TargetType string
	Engine     rendering.Engine
}

type conversionStepResult struct {
	step ConversionStep
	err  error
}

type ConversionPipeline struct {
	fs    *rendering.FS
	steps []ConversionStep
}

func (p *ConversionPipeline) Render(
	ctx context.Context,
	w rendering.DocumentWriter,
	r rendering.DocumentReader,
) error {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	if len(p.steps) == 0 {
		return fmt.Errorf("cannot render: %w", ErrEmptyPipeline)
	}

	results := make(chan conversionStepResult, len(p.steps))

	runStep := func(step ConversionStep, w rendering.DocumentWriter, r rendering.DocumentReader) {
		timer := utils.NewVecTimer(renderDocumentDuration)

		w.SetType(step.TargetType)
		result := conversionStepResult{
			step: step,
			err:  step.Engine.Render(ctx, p.fs, w, r, step.TargetType),
		}
		status := "success"

		if result.err != nil {
			status = "failure"
		}

		if err := w.Close(); err != nil {
			if result.err == nil {
				result.err = err
				status = "failure"
			}
		}

		timer.ObserveDurationWithLabelValues(
			step.SourceType, step.TargetType, step.Engine.Info().Type, status)
		results <- result
	}

	for _, step := range p.steps[:len(p.steps)-1] {
		pw, pr := rendering.DocumentPipe()
		go runStep(step, pw, r)
		r = pr
	}

	go runStep(p.steps[len(p.steps)-1], w, r)

	var err error

	for i := 0; i != len(p.steps); i++ {
		stepResult := <-results
		if stepResult.err != nil {
			if err == nil {
				err = fmt.Errorf(
					"engine %s failed: %w",
					stepResult.step.Engine.Info().Type, stepResult.err)

				cancel()
			}
		}
	}

	return err
}

var (
	renderDocumentDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "rednerd_render_document",
			Help:       "Render document duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{
			"from_type",
			"to_type",
			"engine",
			"status",
		},
	)

	paths = [][]string{
		{
			"text/mjml",
			"text/html",
			"application/pdf",
		},
		{
			"text/mjml",
			"text/html",
			"application/pdf",
			"image/png",
		},
		{
			"text/plain",
			"application/pdf",
			"image/png",
		},
		{
			"application/vnd.oasis.opendocument.text",
			"application/pdf",
			"image/png",
		},
		{
			"application/msword",
			"application/pdf",
			"image/png",
		},
		{
			"application/vnd.openxmlformats-officedocument.wordprocessingml.document",
			"application/pdf",
			"image/png",
		},
		{
			"text/html",
			"application/pdf",
			"image/png",
		},
	}
)

func init() {
	prometheus.MustRegister(renderDocumentDuration)
}
