package redner

import (
	"context"

	"github.com/prometheus/client_golang/prometheus"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/utils"
)

func NewTemplatePipeline(
	ctx context.Context,
	templateRegistry *rendering.TemplateEngineRegistry,
	registry *rendering.EngineRegistry, fs *rendering.FS,
	template *models.Template, metadata models.Metadata, targetType string,
) (*TemplatePipeline, error) {
	tmpl, err := templateRegistry.Load(ctx, fs, template)
	if err != nil {
		return nil, err
	}

	var conversionPipeline *ConversionPipeline

	if template.Produces != targetType {
		p, err := NewConversionPipeline(registry, fs, template.Produces, targetType)
		if err != nil {
			return nil, err
		}

		conversionPipeline = p
	}

	return &TemplatePipeline{
		templateLanguage: template.Language,
		templateProduces: template.Produces,
		metadata:         metadata,
		template:         tmpl,
		pipeline:         conversionPipeline,
	}, nil
}

type TemplatePipeline struct {
	templateLanguage string
	templateProduces string
	metadata         models.Metadata
	template         rendering.Template
	pipeline         *ConversionPipeline
}

func (t *TemplatePipeline) Render(
	ctx context.Context, w rendering.DocumentWriter, record models.Record,
) error {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	pipeErr := make(chan error, 1)

	if t.pipeline != nil {
		pw, pr := rendering.DocumentPipe()
		go func(ctx context.Context, w rendering.DocumentWriter, r rendering.DocumentReader) {
			pipeErr <- t.pipeline.Render(ctx, w, r)
		}(ctx, w, pr)

		w = pw
	}

	w.SetType(t.templateProduces)

	templateTimer := utils.NewVecTimer(renderTemplateDuration)

	if err := t.template.Render(ctx, w, record); err != nil {
		_ = w.Close()

		templateTimer.ObserveDurationWithLabelValues(t.templateLanguage, t.templateProduces, "failure")

		return err
	}

	templateTimer.ObserveDurationWithLabelValues(t.templateLanguage, t.templateProduces, "success")

	if err := w.Close(); err != nil {
		return err
	}

	if t.pipeline != nil {
		return <-pipeErr
	}

	return nil
}

var renderTemplateDuration = prometheus.NewSummaryVec(
	prometheus.SummaryOpts{
		Name:       "rednerd_render_template",
		Help:       "Render template duration and count",
		Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
	},
	[]string{
		"template_language",
		"to_type",
		"status",
	},
)

func init() {
	prometheus.MustRegister(renderTemplateDuration)
}
