package models

import (
	"bytes"
	"crypto/sha256"
	"encoding/hex"
	"io"
	"slices"
	"strings"

	"github.com/Masterminds/squirrel"
	"orus.io/orus-io/go-orusapi/database"
)

// dbtable:"fsnode"
type FSNode struct {
	Root string `db:"root"`
	Path string `db:"path"`

	Mimetype string `db:"mimetype"`
	Size     int64  `db:"size"`
	Sha256   []byte `db:"sha256"`
}

func (node *FSNode) Filename() string {
	segments := strings.Split(node.Path, "/")

	return segments[len(segments)-1]
}

func (node *FSNode) GetBody(db *database.SQLHelper) (io.ReadCloser, error) {
	var data []byte

	if err := db.Get(
		&data, squirrel.Select("body").From("fscontent").Where(squirrel.Eq{"sha256": node.Sha256}),
	); err != nil {
		return nil, err
	}

	return io.NopCloser(bytes.NewReader(data)), nil
}

func GetFileBody(db *database.SQLHelper, sha256Hex string) (io.ReadCloser, error) {
	sha, err := hex.DecodeString(sha256Hex)
	if err != nil {
		return nil, err
	}

	var data []byte

	if err := db.Get(
		&data, squirrel.Select("body").From("fscontent").Where(squirrel.Eq{"sha256": sha}),
	); err != nil {
		return nil, err
	}

	return io.NopCloser(bytes.NewReader(data)), nil
}

func InsertFile(db *database.SQLHelper, node *FSNode, body []byte) error {
	node.Path = "/" + strings.Trim(node.Path, "/")

	node.Size = int64(len(body))

	sha := sha256.Sum256(body)
	node.Sha256 = sha[:]

	if _, err := db.Exec(
		squirrel.Insert(FSNodeTable).
			Columns(FSNodeColumns...).
			Values(node.Values(FSNodeColumns...)...),
	); err != nil {
		return err
	}

	segments := slices.DeleteFunc(
		strings.Split(node.Path, "/"),
		func(v string) bool { return v == "" })

	parents := make([]string, len(segments)+1)

	for i := range segments {
		parents[i+1] = "/" + strings.Join(segments[:i+1], "/")
	}

	parents[0] = "/"

	q := squirrel.Insert("fsclosure").
		Columns("root", "parent_path", "child_path")

	for _, p := range parents {
		q = q.Values(
			node.Root, p, node.Path,
		)
	}

	if _, err := db.Exec(q); err != nil {
		return err
	}

	// TODO avoid sending data if already exist
	if _, err := db.ExecRaw(
		`INSERT INTO fscontent VALUES ($1, $2) ON CONFLICT DO NOTHING`, sha[:], body,
	); err != nil {
		return err
	}

	return nil
}

func DeleteFile(db *database.SQLHelper, root string, path string) error {
	sql, args, err := squirrel.Select("child_path").
		From("fsclosure").
		Where(squirrel.Eq{
			"root":        root,
			"parent_path": path,
		}).ToSql()
	if err != nil {
		panic(err)
	}

	q := squirrel.
		Delete(FSNodeTable).
		Prefix("WITH paths AS ( "+sql+" ) ", args...).
		Where(squirrel.Expr("path IN (SELECT child_path FROM paths)")).
		Where(squirrel.Eq{
			"root": root,
		})

	if _, err := db.Exec(q); err != nil {
		return err
	}

	return nil
}

func ListFiles(db *database.SQLHelper, root string, path string) ([]FSNode, error) {
	var files []FSNode
	if err := db.Select(
		&files,
		squirrel.
			Select(
				Schema.FSNode.FQColumns(true)...,
			).
			From(Schema.FSNode.GetName()).
			LeftJoin(
				"fsclosure ON fsclosure.root = fsnode.root AND fsclosure.child_path = fsnode.path",
			).
			Where(squirrel.Eq{
				"fsclosure.root":        root,
				"fsclosure.parent_path": path,
			}),
	); err != nil {
		return nil, err
	}

	return files, nil
}
