package rendering

import (
	"bytes"
	"encoding/hex"
	"io"
	"maps"
	"os"
	"path/filepath"
	"slices"
	"strings"

	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
)

type fsLayer struct {
	root  string
	files []*models.File
}

type FS struct {
	db *database.SQLHelper

	layers []fsLayer

	fileIndex map[string]*models.File
	fileRoot  map[string]string
	fileList  []*models.File
}

func NewFS(db *database.SQLHelper) *FS {
	return &FS{
		db:        db,
		fileIndex: make(map[string]*models.File),
		fileRoot:  make(map[string]string),
	}
}

func (fs *FS) AddRoot(root string) error {
	nodes, err := models.ListFiles(fs.db, root, "/")
	if err != nil {
		return err
	}

	files := make([]*models.File, len(nodes))
	for i, node := range nodes {
		files[i] = &models.File{
			Path:   node.Path,
			Type:   node.Mimetype,
			Sha256: hex.EncodeToString(node.Sha256),
			Size:   node.Size,
		}
	}

	fs.addLayer(root, files)

	return nil
}

func (fs *FS) AddLayer(files []*models.File) {
	fs.addLayer("", files)
}

func (fs *FS) addLayer(root string, files []*models.File) {
	for _, f := range files {
		if f.Type == "redner/mask" {
			for k := range maps.Keys(fs.fileIndex) {
				if k == f.Path || strings.HasPrefix(k, f.Path+"/") {
					delete(fs.fileIndex, k)
					delete(fs.fileRoot, k)
				}
			}
		}
	}

	for _, f := range files {
		if !strings.HasPrefix(f.Type, "redner/") {
			fs.fileRoot[f.Path] = root
			fs.fileIndex[f.Path] = f
		}
	}

	fs.layers = append(fs.layers, fsLayer{
		root:  root,
		files: files,
	})

	fs.fileList = make([]*models.File, 0, len(fs.fileIndex))
	for _, v := range fs.fileIndex {
		fs.fileList = append(fs.fileList, v)
	}

	slices.SortFunc(fs.fileList, func(a, b *models.File) int {
		return strings.Compare(a.Path, b.Path)
	})
}

func (fs *FS) List() []*models.File {
	return fs.fileList
}

func normalizePath(path string) string {
	return "/" + strings.Join(slices.DeleteFunc(
		strings.Split(path, "/"),
		func(p string) bool {
			return p == "" || p == "."
		}), "/")
}

func (fs *FS) ListDir(path string) []*models.File {
	path = normalizePath(path)
	if !strings.HasSuffix(path, "/") {
		path += "/"
	}

	var list []*models.File

	for _, f := range fs.fileList {
		if strings.HasPrefix(f.Path, path) {
			list = append(list, f)
		}
	}

	return list
}

func fileExists(path string) bool {
	_, err := os.Stat(path)

	return err == nil
}

func (fs *FS) SyncTo(cacheDir string, target string, paths ...string) error {
	var filelist []*models.File

	for _, p := range paths {
		filelist = append(filelist, fs.ListDir(p)...)
	}

	workdir, err := os.MkdirTemp("", "")
	if err != nil {
		return err
	}
	defer os.RemoveAll(workdir)

	for _, f := range filelist {
		wPath := filepath.Join(workdir, f.Path)

		cachePath := filepath.Join(cacheDir, f.Sha256)
		if !fileExists(cachePath) {
			tgt, err := os.CreateTemp(cacheDir, "tmp")
			if err != nil {
				return err
			}

			reader, err := fs.Read(f.Path)
			if err != nil {
				return err
			}

			if _, err := io.Copy(tgt, reader); err != nil {
				return err
			}

			if err := reader.Close(); err != nil {
				return err
			}

			if err := tgt.Close(); err != nil {
				return err
			}

			if err := os.Rename(tgt.Name(), cachePath); err != nil {
				return err
			}
		}

		if err := os.MkdirAll(filepath.Dir(wPath), os.ModePerm); err != nil {
			return err
		}

		if err := os.Link(cachePath, wPath); err != nil {
			return err
		}
	}

	for _, p := range paths {
		if err := os.RemoveAll(filepath.Join(target, p)); err != nil {
			return err
		}

		wdir := filepath.Join(workdir, p)
		if fileExists(wdir) {
			if err := os.Rename(wdir, filepath.Join(target, p)); err != nil {
				return err
			}
		}
	}

	return nil
}

func (fs *FS) Read(path string) (io.ReadCloser, error) {
	f, ok := fs.fileIndex[path]
	if !ok {
		return nil, os.ErrNotExist
	}

	if f.Body != nil {
		return io.NopCloser(bytes.NewReader(f.Body)), nil
	}

	if fs.fileRoot[path] == "" {
		return nil, os.ErrNotExist
	}

	reader, err := models.GetFileBody(fs.db, f.Sha256)
	if err != nil {
		return nil, err
	}

	if f.Size < 10*1024*1024 {
		return &fileReader{
			ReadCloser: reader,
			target:     f,
		}, nil
	}

	return reader, nil
}

type fileReader struct {
	io.ReadCloser
	buf    []byte
	target *models.File
}

func (r *fileReader) Read(data []byte) (int, error) {
	n, err := r.ReadCloser.Read(data)

	r.buf = append(r.buf, data[:n]...)

	if err == io.EOF {
		r.target.Body = r.buf
	}

	return n, err
}
