package rendering_test

import (
	"crypto/sha256"
	"encoding/hex"
	sfs "io/fs"
	"os"
	"path/filepath"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/go-orusapi/database"
	"orus.io/orus-io/go-orusapi/testutils"

	"orus.io/orus-io/rednerd/migration"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
)

func TestFS(t *testing.T) {
	t.Run("SyncTo", func(t *testing.T) {
		log := testutils.GetLogger(t)
		testdb := database.GetTestDB(t.Context(), t, migration.Source)
		db := database.NewSQLHelper(t.Context(), testdb, log)

		filedata1 := []byte("some content")
		filedata1Sum := sha256.Sum256(filedata1)
		filedata2 := []byte("some other content")
		filedata2Sum := sha256.Sum256(filedata2)
		filedata3 := []byte("yet another content")
		filedata3Sum := sha256.Sum256(filedata3)
		filedata4 := []byte("yet another content, again")
		filedata4Sum := sha256.Sum256(filedata4)

		fs := rendering.NewFS(&db)
		fs.AddLayer([]*models.File{{
			Path:   "/fonts/font1.ttf",
			Type:   "font/ttf",
			Sha256: hex.EncodeToString(filedata1Sum[:]),
			Body:   filedata1,
		}, {
			Path:   "/fonts/font2.ttf",
			Type:   "font/ttf",
			Sha256: hex.EncodeToString(filedata2Sum[:]),
			Body:   filedata2,
		}, {
			Path:   "/modules/mod1/mod.json",
			Type:   "application/json",
			Sha256: hex.EncodeToString(filedata3Sum[:]),
			Body:   filedata3,
		}, {
			Path:   "/modules/mod2/mod.json",
			Type:   "application/json",
			Sha256: hex.EncodeToString(filedata4Sum[:]),
			Body:   filedata4,
		}})

		dir := t.TempDir()

		t.Cleanup(func() {
			t.Log("Dir content:")
			assert.NoError(t, sfs.WalkDir(os.DirFS(dir), ".", func(path string, d sfs.DirEntry, err error) error {
				t.Log(path)

				return nil
			}))
		})

		cacheDir := filepath.Join(dir, "cache")
		targetDir := filepath.Join(dir, "target")

		require.NoError(t, os.Mkdir(cacheDir, os.ModePerm))
		require.NoError(t, os.Mkdir(targetDir, os.ModePerm))

		t.Run("fonts-dir", func(t *testing.T) {
			t.Cleanup(func() {
				require.NoError(t, os.RemoveAll(targetDir))
				require.NoError(t, os.Mkdir(targetDir, os.ModePerm))
			})

			assert.NoError(t, fs.SyncTo(cacheDir, targetDir, "fonts"))

			assert.DirExists(t, filepath.Join(targetDir, "fonts"))
			assert.FileExists(t, filepath.Join(targetDir, "fonts/font1.ttf"))
			assert.FileExists(t, filepath.Join(targetDir, "fonts/font2.ttf"))
			assert.NoDirExists(t, filepath.Join(targetDir, "modules"))
		})

		t.Run("dir-with-subdirs", func(t *testing.T) {
			t.Cleanup(func() {
				require.NoError(t, os.RemoveAll(targetDir))
				require.NoError(t, os.Mkdir(targetDir, os.ModePerm))
			})

			assert.NoError(t, fs.SyncTo(cacheDir, targetDir, "modules"))

			assert.NoDirExists(t, filepath.Join(targetDir, "fonts"))
			assert.DirExists(t, filepath.Join(targetDir, "modules"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod1/mod.json"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod2/mod.json"))

			assert.NoError(t, fs.SyncTo(cacheDir, targetDir, "modules"))

			assert.NoDirExists(t, filepath.Join(targetDir, "fonts"))
			assert.DirExists(t, filepath.Join(targetDir, "modules"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod1/mod.json"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod2/mod.json"))
		})

		t.Run("multi-dir", func(t *testing.T) {
			t.Cleanup(func() {
				require.NoError(t, os.RemoveAll(targetDir))
				require.NoError(t, os.Mkdir(targetDir, os.ModePerm))
			})

			assert.NoError(t, fs.SyncTo(cacheDir, targetDir, "fonts", "modules"))

			assert.FileExists(t, filepath.Join(targetDir, "fonts/font1.ttf"))
			assert.FileExists(t, filepath.Join(targetDir, "fonts/font2.ttf"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod1/mod.json"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod2/mod.json"))
		})

		t.Run("full", func(t *testing.T) {
			t.Cleanup(func() {
				require.NoError(t, os.RemoveAll(targetDir))
				require.NoError(t, os.Mkdir(targetDir, os.ModePerm))
			})

			assert.NoError(t, fs.SyncTo(cacheDir, targetDir, "."))

			assert.FileExists(t, filepath.Join(targetDir, "fonts/font1.ttf"))
			assert.FileExists(t, filepath.Join(targetDir, "fonts/font2.ttf"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod1/mod.json"))
			assert.FileExists(t, filepath.Join(targetDir, "modules/mod2/mod.json"))
		})
	})
}
