package rendering

import (
	"context"
	"errors"
	"maps"
	"slices"
	"strings"

	"orus.io/orus-io/rednerd/models"
)

// ErrUnknownTemplateLanguage is returned if no engine supports the given language.
var ErrUnknownTemplateLanguage = errors.New("unknown template language")

// TemplateEngineInfo describes an engine capabilities.
type TemplateEngineInfo struct {
	Type        string
	Language    string
	Produces    []string
	Description string
	Metadata    []EngineMetadataInfo
}

// Template is a parsed and loaded template, ready to be rendered.
type Template interface {
	Varlist() ([]string, error)
	Render(ctx context.Context, w DocumentWriter, record models.Record) error
}

// TemplateEngine is able to render a document.
type TemplateEngine interface {
	Info() TemplateEngineInfo
	Load(ctx context.Context, fs *FS, template *models.Template) (Template, error)
}

// RegisterTemplateEngine a new engine factory.
func RegisterTemplateEngine(name string, engine TemplateEngine) error {
	return DefaultTemplateEngineRegistry.Register(name, engine)
}

// MustRegisterTemplateEngine registers an TemplateEngine and panics if it fails.
func MustRegisterTemplateEngine(name string, engine TemplateEngine) {
	DefaultTemplateEngineRegistry.MustRegister(name, engine)
}

// TemplateEngineRegistry is a set of TemplateEngine, ready for use.
type TemplateEngineRegistry struct {
	templateEngines map[string]TemplateEngine
}

func (r *TemplateEngineRegistry) All() map[string]TemplateEngine {
	return maps.Clone(r.templateEngines)
}

// Register a TemplateEngine.
func (r *TemplateEngineRegistry) Register(name string, engine TemplateEngine) error {
	if _, ok := r.templateEngines[name]; ok {
		return ErrDuplicateName
	}

	r.templateEngines[name] = engine

	return nil
}

// MustRegister a TemplateEngine.
func (r *TemplateEngineRegistry) MustRegister(name string, engine TemplateEngine) {
	if err := r.Register(name, engine); err != nil {
		panic(err)
	}
}

func (r *TemplateEngineRegistry) find(language string, produces string) TemplateEngine {
	for _, engine := range r.templateEngines {
		if engine.Info().Language == language {
			if (produces == "" && len(engine.Info().Produces) == 0) ||
				slices.Contains(engine.Info().Produces, produces) {
				return engine
			}
		}
	}

	return nil
}

func mimelookuplist(v string) []string {
	base, _, _ := strings.Cut(v, ";")
	mediatype := strings.TrimSpace(strings.ToLower(base))

	list := []string{"", "*/*"}

	t, st, ok := strings.Cut(mediatype, "/")
	if ok {
		if t != "*" {
			list = append(list, t+"/*")
		}

		if st != "*" {
			list = append(list, t+"/"+st)
		}
	}

	slices.Reverse(list)

	return list
}

// Find finds a template engine by the language it supports.
func (r *TemplateEngineRegistry) Find(language string, produces string) TemplateEngine {
	for _, mimetype := range mimelookuplist(produces) {
		if engine := r.find(language, mimetype); engine != nil {
			return engine
		}
	}

	return nil
}

// Load a template with the proper engine.
func (r *TemplateEngineRegistry) Load(ctx context.Context, fs *FS, template *models.Template) (Template, error) {
	engine := r.Find(template.Language, template.Produces)
	if engine == nil {
		return nil, ErrUnknownTemplateLanguage
	}

	return engine.Load(ctx, fs, template)
}

// NewTemplateEngineRegistry creates a TemplateEngineRegistry.
func NewTemplateEngineRegistry() *TemplateEngineRegistry {
	return &TemplateEngineRegistry{
		templateEngines: make(map[string]TemplateEngine),
	}
}

// DefaultTemplateEngineRegistry is the default engine registry.
var DefaultTemplateEngineRegistry = NewTemplateEngineRegistry()
