package handlers

import (
	"errors"
	"fmt"
	"net/http"

	"github.com/go-openapi/runtime"
	"github.com/go-openapi/runtime/middleware"
	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/utils"
)

var (
	// ErrNoSuchTemplate is returned if the template does not exist.
	ErrNoSuchTemplate = errors.New("no such template")
	// ErrDuplicateTemplateName is returned if the template's name already exists.
	ErrDuplicateTemplateName = errors.New("duplicate template name")
	// ErrNoTemplateNorDocuments is returned if the render request has no template
	// nor a document.
	ErrNoTemplateNorDocument = errors.New("no template nor document. Must have one")
	// ErrNoTemplate is returned if the template-varlist request has no template.
	ErrNoTemplate = errors.New("no template given")
	// ErrUnknownTemplateLanguage is returned when the language is unknown.
	ErrUnknownTemplateLanguage = errors.New("unknown template language")
	// ErrEmptyField is returned when a mandatory field is empty.
	ErrEmptyField = errors.New("empty field")
	// ErrNoSuchUser is returned if the user does not exist.
	ErrNoSuchUser = errors.New("no such user")
	// ErrInvalidAuth is returned if the principal is nil.
	ErrInvalidAuth = errors.New("invalid auth")
	// ErrRenderedDocumentSizeMismatch is returned when there is an expected
	// Number of documents that /render must return (like when previewing).
	ErrRenderedDocumentSizeMismatch = errors.New("rendered documents size mismatch")
)

func NotFound(msg string) utils.HTTPError {
	return utils.NewHTTPError(http.StatusNotFound, errors.New(msg)) //nolint:err113
}

func NotFoundf(msg string, args ...any) utils.HTTPError {
	return utils.NewHTTPErrorf(http.StatusNotFound, msg, args...)
}

func HandleError(err error, log *zerolog.Logger) middleware.Responder {
	var httpErr utils.HTTPError
	if errors.As(err, &httpErr) {
		return &HTTPErrorResponder{httpErr}
	}

	return InternalError(err, log, false)
}

// InternalError is a helper function to log and internal error and obtain
// a Responder that you can use as a return value in your handler.
func InternalError(err error, log *zerolog.Logger, debug bool) *InternalErrorResponder {
	log.Err(err).Msg("internal error")

	return &InternalErrorResponder{err: err, log: log, debug: debug}
}

func BadRequest(msg string) *HTTPErrorResponder {
	return &HTTPErrorResponder{utils.HTTPBadRequest(fmt.Errorf("%s", msg))} //nolint:err113
}

func BadRequestError(err error) *HTTPErrorResponder {
	return &HTTPErrorResponder{utils.HTTPBadRequest(err)}
}

func Unauthorized(msg string) *HTTPErrorResponder {
	return &HTTPErrorResponder{
		utils.NewHTTPErrorf(http.StatusUnauthorized, "%s", msg),
	}
}

func Unauthorizedf(msg string, args ...any) *HTTPErrorResponder {
	return &HTTPErrorResponder{
		utils.NewHTTPErrorf(http.StatusUnauthorized, msg, args...),
	}
}

// InternalErrorResponder ...
// InternalErrorResponder is a helper to filter out the error content sent
// to the client this filtering is dependent on the global debug flag.
// ie: if debug is set then the client will receive the detail
// of the internal error.
// This is something you obviously do NOT want to do in production mode.
type InternalErrorResponder struct {
	err   error
	log   *zerolog.Logger
	debug bool
}

// WriteResponse to the client.
func (r *InternalErrorResponder) WriteResponse(rw http.ResponseWriter, producer runtime.Producer) {
	rw.WriteHeader(http.StatusInternalServerError)

	if r.err != nil && r.debug {
		// clients are allowed to see internal errors content only if the
		// global debug mode is activated on the ALPE api
		if err := producer.Produce(rw, models.FromError(r.err)); err != nil {
			panic(err) // let the recovery middleware deal with this
		}
	}
}

func NewHTTPErrorResponder(err error) *HTTPErrorResponder {
	var httpErr utils.HTTPError
	if !errors.As(err, &httpErr) {
		httpErr = utils.NewHTTPError(http.StatusInternalServerError, err)
	}

	return &HTTPErrorResponder{
		err: httpErr,
	}
}

// HTTPErrorResponderError is a generic http error responder.
type HTTPErrorResponder struct { //nolint:errname
	err utils.HTTPError
}

func (r HTTPErrorResponder) Error() string {
	return "BadRequest" + r.err.Error()
}

func (r HTTPErrorResponder) Unwrap() error {
	return r.err
}

// WriteResponse to the client.
func (r HTTPErrorResponder) WriteResponse(rw http.ResponseWriter, producer runtime.Producer) {
	rw.WriteHeader(r.err.StatusCode)

	if err := producer.Produce(rw, models.FromError(r.err.Err)); err != nil {
		panic(err)
	}
}
