package handlers

import (
	"encoding/hex"

	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/file_management"
)

func NewFileCreateHandler(db *sqlx.DB) *FileCreateHandler {
	return &FileCreateHandler{db}
}

type FileCreateHandler struct {
	db *sqlx.DB
}

func (h *FileCreateHandler) Handle(
	params file_management.FileCreateParams,
	principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)

	var root string

	if params.Root != nil {
		root = *params.Root
	}

	perms, err := GetFileRootPerms(root, principal)
	if err != nil {
		return BadRequestError(err)
	}

	if !perms.Create {
		return Unauthorized("")
	}

	path, err := NormalizeFilePath(params.File.Path)
	if err != nil {
		return NewHTTPErrorResponder(err)
	}

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return InternalError(err, log, false)
	}
	defer tx.RollbackIfOpened(*log)

	db := database.NewSQLHelper(ctx, tx, *log)

	node := models.FSNode{
		Root:     root,
		Path:     path,
		Mimetype: params.File.Type,
	}
	if err := models.InsertFile(
		&db, &node, params.File.Body,
	); err != nil {
		return InternalError(err, log, false)
	}

	if err := tx.Commit(); err != nil {
		return InternalError(err, log, false)
	}

	return file_management.NewFileCreateCreated().WithPayload(
		&models.File{
			Path:   node.Path,
			Type:   node.Mimetype,
			Sha256: hex.Dump(node.Sha256),
			Size:   node.Size,
		},
	)
}
