package handlers

import (
	"database/sql"
	"encoding/hex"
	"errors"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/file_management"
)

func NewFileDownloadHandler(db *sqlx.DB) *FileDownloadHandler {
	return &FileDownloadHandler{db}
}

type FileDownloadHandler struct {
	db *sqlx.DB
}

func (h *FileDownloadHandler) Handle(
	params file_management.FileDownloadParams,
	principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)

	var root string

	if params.Root != nil {
		root = *params.Root
	}

	path, err := NormalizeFilePath(params.Path)
	if err != nil {
		return NewHTTPErrorResponder(err)
	}

	perms, err := GetFileRootPerms(root, principal)
	if err != nil {
		return BadRequestError(err)
	}

	if !perms.Read {
		return Unauthorized("")
	}

	db := database.NewSQLHelper(ctx, h.db, *log)

	var node models.FSNode

	if err := db.Get(&node, squirrel.Select(models.FSNodeColumns...).
		From(models.FSNodeTable).
		Where(squirrel.Eq{
			models.FSNodeRootColumn: root,
			models.FSNodePathColumn: path,
		}),
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return file_management.NewFileDownloadNotFound()
		}

		return InternalError(err, log, false)
	}

	reader, err := node.GetBody(&db)
	if err != nil {
		return InternalError(err, log, false)
	}

	return NewFileResponse().
		WithFilename(node.Filename()).
		WithMimetype(node.Mimetype).
		WithETag(hex.Dump(node.Sha256)).
		WithBody(reader)
}
