package handlers

import (
	"errors"
	"fmt"
	"slices"
	"strings"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/file_management"
)

// NormalizeFilePath returns an error if the file path is not valid (contains
// "." or "..".
func NormalizeFilePath(path string) (string, error) {
	segments := slices.DeleteFunc(
		strings.Split(path, "/"),
		func(v string) bool { return v == "" || v == "." })

	if slices.Contains(segments, "..") {
		return "", BadRequest("'..' is not allowed in file paths")
	}

	return "/" + strings.Join(segments, "/"), nil
}

type FileRootPermissions struct {
	Read   bool
	Create bool
	Delete bool
}

var ErrInvalidRoot = errors.New("invalid root")

func GetFileRootPerms(
	root string, principal *models.Principal,
) (FileRootPermissions, error) {
	if root == "global" {
		return FileRootPermissions{
			Read:   true,
			Create: principal.IsAdmin,
			Delete: principal.IsAdmin,
		}, nil
	}

	if id, ok := strings.CutPrefix(root, "account/"); ok {
		if principal.IsAdmin {
			return FileRootPermissions{
				Read:   false,
				Create: false,
				Delete: false,
			}, nil
		}

		accountMatch := id == principal.Subject

		return FileRootPermissions{
			Read:   accountMatch,
			Create: accountMatch,
			Delete: accountMatch,
		}, nil
	}

	if id, ok := strings.CutPrefix(root, "template/"); ok {
		if principal.IsAdmin {
			return FileRootPermissions{
				Read:   false,
				Create: false,
				Delete: false,
			}, nil
		}

		if account, _, ok := strings.Cut(id, "/"); ok {
			accountMatch := account == principal.Subject

			return FileRootPermissions{
				Read:   accountMatch,
				Create: accountMatch,
				Delete: accountMatch,
			}, nil
		}

		return FileRootPermissions{}, fmt.Errorf("%w: %s", ErrInvalidRoot, root)
	}

	return FileRootPermissions{
		Read:   false,
		Create: false,
		Delete: false,
	}, nil
}

func NewFileListHandler(db *sqlx.DB) *FileListHandler {
	return &FileListHandler{db}
}

type FileListHandler struct {
	db *sqlx.DB
}

func (h *FileListHandler) Handle(
	params file_management.FileListParams,
	principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)

	var (
		root string
		path string
	)

	if params.Root != nil {
		root = *params.Root
	}

	if params.Path != nil {
		path = *params.Path
	}

	path, err := NormalizeFilePath(path)
	if err != nil {
		return NewHTTPErrorResponder(err)
	}

	perms, err := GetFileRootPerms(root, principal)
	if err != nil {
		return BadRequestError(err)
	}

	if !perms.Read {
		return Unauthorized("")
	}

	db := database.NewSQLHelper(ctx, h.db, *log)

	files, err := h.FileList(&db, root, path)
	if err != nil {
		return HandleError(err, log)
	}

	return file_management.NewFileListOK().WithPayload(files)
}

func (h *FileListHandler) FileList(
	db *database.SQLHelper, root string, path string,
) (*models.Fileset, error) {
	rows, err := db.Query(
		squirrel.
			Select(
				models.Schema.FSNode.FQColumns(true)...,
			).
			From(models.FSNodeTable).
			LeftJoin(
				"fsclosure ON fsclosure.root = fsnode.root AND fsclosure.child_path = fsnode.path",
			).
			Where(squirrel.Eq{
				"fsclosure.root":        root,
				"fsclosure.parent_path": path,
			}).
			OrderBy(models.Schema.FSNode.Path.Name()),
	)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var fileset models.Fileset

	for rows.Next() {
		var node models.FSNode

		if err := rows.StructScan(&node); err != nil {
			return nil, err
		}

		fileset.Files = append(fileset.Files, &models.File{
			Path:   node.Path,
			Type:   node.Mimetype,
			Size:   node.Size,
			Sha256: string(node.Sha256),
		})
	}

	return &fileset, nil
}
