package tests

import (
	"encoding/base64"
	"net/http"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils/apitester"
)

func TestFileManagementAPI(t *testing.T) {
	tester := apitester.NewAPITester(t, nil)
	defer tester.Close()

	defer tester.Logout()

	tester.Login("admin", "admin")

	tester.CreateUser("janedoe", "janedoe")
	tester.Login("janedoe", "janedoe")

	var template models.Template

	tester.CreateTemplate(
		"janedoe",
		&models.Template{
			Name:       "t1",
			Language:   "mustache",
			Produces:   "text/html",
			BodyFormat: "text",
			Body:       "<html><body>Hi {{name}}!</body></html>",
		}, &template)

	t.Run("create_global_files", func(t *testing.T) {
		defer tester.SetT(t)()

		tester.Login("admin", "admin")

		var file models.File

		filecontent := base64.StdEncoding.EncodeToString([]byte("Hello world !"))
		tester.APITest("create_global_file_1").
			Post("/api/v1/file?root=global").
			JSON(`{"path":"/dir1/file1.txt","type":"text/plain","body":"` + filecontent + `"}`).
			Expect(t).
			Status(http.StatusCreated).
			End().
			JSON(&file)

		tester.APITest("create_global_file_2").
			Post("/api/v1/file?root=global").
			JSON(`{"path":"/dir1/dir2/file2.txt","type":"text/plain","body":"` + filecontent + `"}`).
			Expect(t).
			Status(http.StatusCreated).
			End().
			JSON(&file)
	})

	t.Run("read_global_files", func(t *testing.T) {
		defer tester.SetT(t)()

		var files models.Fileset

		tester.APITest("read_global_files").
			Get("/api/v1/file").
			Expect(t).
			Status(http.StatusOK).
			End().
			JSON(&files)

		require.Len(t, files.Files, 2)
		assert.Equal(t, "/dir1/dir2/file2.txt", files.Files[0].Path)
		assert.Equal(t, "/dir1/file1.txt", files.Files[1].Path)

		tester.APITest("read_global_files").
			Get("/api/v1/file").
			Query("root", "global").
			Query("path", "/dir1/dir2").
			Expect(t).
			Status(http.StatusOK).
			End().
			JSON(&files)

		require.Len(t, files.Files, 1)
		assert.Equal(t, "/dir1/dir2/file2.txt", files.Files[0].Path)
	})

	t.Run("download_global_files", func(t *testing.T) {
		defer tester.SetT(t)()

		tester.APITest("download_global_files").
			Get("/api/v1/file/download").
			Query("path", "dir1/dir2/file2.txt").
			Expect(t).
			Status(http.StatusOK).
			Header("Content-Type", "text/plain").
			Header("Content-Disposition", "attachment; filename=\"file2.txt\"").
			Body("Hello world !").
			End()
	})

	t.Run("delete_global_files", func(t *testing.T) {
		defer tester.SetT(t)()

		tester.Login("janedoe", "janedoe")

		tester.APITest("delete_global_files").
			Delete("/api/v1/file").
			Query("path", "/dir1/dir2").
			Expect(t).
			Status(http.StatusUnauthorized).
			End()

		tester.Login("admin", "admin")

		tester.APITest("delete_global_files").
			Delete("/api/v1/file").
			Query("path", "/dir1/dir2").
			Expect(t).
			Status(http.StatusNoContent).
			End()

		var files models.Fileset

		tester.APITest("read_global_files").
			Get("/api/v1/file").
			Query("root", "global").
			Query("path", "/").
			Expect(t).
			Status(http.StatusOK).
			End().
			JSON(&files)

		require.Len(t, files.Files, 1)

		tester.APITest("delete_global_files").
			Delete("/api/v1/file").
			Query("path", "/dir1/file1.txt").
			Expect(t).
			Status(http.StatusNoContent).
			End()

		tester.APITest("read_global_files").
			Get("/api/v1/file").
			Query("root", "global").
			Query("path", "/").
			Expect(t).
			Status(http.StatusOK).
			End().
			JSON(&files)

		require.Len(t, files.Files, 0)
	})
}
